#ifndef VINS_MOBILE_ANDROIDPORT_VIEWCONTROLLER_H
#define VINS_MOBILE_ANDROIDPORT_VIEWCONTROLLER_H

#include "opencv2/imgproc/types_c.h"
#include "utility.hpp"
#include "feature_tracker.hpp"

#include "global_param.hpp"
#include "VINS.hpp"
#include <queue>
#include "draw_result.hpp"

#include "keyframe.h"
#include "loop_closure.h"
#include "keyfame_database.h"
#include <sys/utsname.h>

#include <condition_variable>
#include <android/log.h>
#include <thread>
#include <jni.h>
#include <android/looper.h>
#include <android/sensor.h>
extern "C" {
    #include <time.h>
}

#define APPNAME "VINS_AR"
#define LOG_TAG "VINSController.cpp"
#define LOGI(...)  __android_log_print(ANDROID_LOG_INFO, LOG_TAG, __VA_ARGS__)
#define LOGD(...)  __android_log_print(ANDROID_LOG_DEBUG, LOG_TAG, __VA_ARGS__)
#define LOGE(...)  __android_log_print(ANDROID_LOG_ERROR, LOG_TAG, __VA_ARGS__)
#define printf(...)  __android_log_print(ANDROID_LOG_ERROR, LOG_TAG, __VA_ARGS__)

typedef double NSTimeInterval;

struct IMU_MSG {
    NSTimeInterval header;
    Vector3d acc;
    Vector3d gyr;
};

struct IMG_MSG {
    NSTimeInterval header;
    map<int, Vector3d> point_clouds;
};

struct IMG_DATA {
    NSTimeInterval header;
    cv::Mat image; // UIImage *image;
};

struct IMG_DATA_CACHE {
    NSTimeInterval header;
    cv::Mat equ_image;
    cv::Mat image; // UIImage *image;
};

struct VINS_DATA_CACHE {
    NSTimeInterval header;
    Vector3f P;
    Matrix3f R;
};

typedef shared_ptr <IMU_MSG const > ImuConstPtr;
typedef shared_ptr <IMG_MSG const > ImgConstPtr;


class VINSController {
private:
    static VINSController* instance;    // singleton for static callbacks

    const int videoWidth = 480;
    const int videoHeight = 640;

    double cur_cliplimit = 3;

    bool isCapturing;
    cv::Ptr<FeatureTracker> feature_tracker;
    cv::Size frameSize;
    std::mutex _condition;
    std::thread mainLoop;
    std::thread drawThread;
    std::thread saveDataThread;
    std::thread loopDetectThread;
    std::thread globalPoseOptmzThread;

    VINS vins;
    FeatureTracker featuretracker;
    queue<ImgConstPtr> img_msg_buf; // Store the fesature data processed by featuretracker
    queue<ImuConstPtr> imu_msg_buf; // Store the IMU data for vins
    queue<IMU_MSG_LOCAL> local_imu_msg_buf; // Store the IMU data for motion-only vins
    int waiting_lists = 0;  // The number of measurements waiting to be processed
    int frame_cnt = 0;
    std::mutex m_buf;   // Lock the feature and imu data buffer
    std::condition_variable con;
    NSTimeInterval current_time = -1;
    NSTimeInterval lateast_imu_time = -1;

    int segmentation_index = 0; // Segment the trajectory using color when re-initialize

    // true:  30 HZ pose output and AR rendering in front-end (very low latency)
    // false: 10 HZ pose output and AR rendering in back-end
    bool USE_PNP = true;

    /** Featuretracker Feedback */
    std::mutex m_depth_feedback;    // Lock the solved VINS data feedback to featuretracker
    std::mutex m_imu_feedback;      // Lock the IMU data feedback to featuretracker
    list<IMG_MSG_LOCAL> solved_features;    // Solved VINS feature feedback to featuretracker
    VINS_RESULT solved_vins;    // Solved VINS status feedback to featuretracker

    /** IMU Things */
    int imu_prepare = 0;
    const int LOOPER_ID_USER = 3;
    const int SENSOR_REFRESH_RATE_HZ = 100;
    const int32_t SENSOR_REFRESH_PERIOD_US = int32_t(1000000 / SENSOR_REFRESH_RATE_HZ);
    static ASensorEventQueue *accelerometerEventQueue;
    static ASensorEventQueue *gyroscopeEventQueue;
    // Sync Gyro & Acc Sensor, then push the latest IMU_MSG to imu_msg_buf and local_imu_msg_buf
    static int process_imu_sensor_events(int fd, int events, void* data);

    /** Loop Closure Things */
    queue<pair<cv::Mat, double>> image_buf_loop;    // Raw image data buffer for extracting FAST feature
    std::mutex m_image_buf_loop;            // Lock the image_buf_loop
    LoopClosure *loop_closure = nullptr;    // Detect loop
    KeyFrameDatabase keyframe_database;     // Loop keyframe database
    int keyframe_freq = 0;                  // Control the loop detection frequency
    int global_frame_cnt = 0;               // Index the keyframe
    int loop_check_cnt = 0;                 // Record the checked loop frame
    bool voc_init_ok = false;               // Indicate if breif vocabulary read finish
    int old_index = -1;                     // Indicate the loop frame index
    Eigen::Vector3d loop_correct_t = Eigen::Vector3d(0, 0, 0);  // Translation drift
    Eigen::Matrix3d loop_correct_r = Eigen::Matrix3d::Identity();   // Rotation drift

    /** UI CONFIG */
    // false:  VINS trajectory is the main view, AR image is in left bottom
    // true: AR image is the main view, VINS is in left bottom
    bool ui_main = false;
    bool box_in_AR = false;
    bool box_in_trajectory = false;
    bool start_show = false;    // True if initialized finished.
    int loop_old_index = -1;
    float x_view_last = -5000;
    float y_view_last = -5000;
    float z_view_last = -5000;
    float total_odom = 0;

    /** Save data for debug */
    queue<IMG_DATA> imgDataBuf;
    IMG_DATA imgData;
    IMU_MSG imuData;
    KEYFRAME_DATA vinsData;
    bool start_record = false;
    bool start_playback = false;
    bool start_playback_vins = false;
    unsigned long imageDataIndex = 0;
    unsigned long imageDataReadIndex = 0;
    unsigned long imuDataIndex = 0;
    unsigned long imuDataReadIndex = 0;
    unsigned long vinsDataIndex = 0;
    unsigned long vinsDataReadIndex = 0;

    // MARK: Unity Camera Mode Switching
    // Ground truth from UI switch property "self.switchUIAREnabled"
    bool imuPredictEnabled = false; // Implied, updated by updateCameraMode()
    bool cameraMode = true; // Implied, updated by updateCameraMode()
    bool imageCacheEnabled = cameraMode && !USE_PNP;    // Implied, updated by updateCameraMode()
    
public:
    VINSController();
    ~VINSController();

    /** View stuff */
    std::mutex viewUpdateMutex;
    std::string tvXText;
    std::string tvYText;
    std::string tvZText;
    std::string tvFeatureText;
    std::string tvTotalText{"TOTAL:"};
    std::string tvBufText;
    std::string tvLoopText{"LOOP:"};
    bool initImageVisible = true;
    float virtualCamDistance = 5;

    /**
     * Main process image thread: this thread detects and track feature between two continuous images
     * and takes the newest VINS result and the corresponding image to draw AR and trajectory.
     * */
    queue<IMG_DATA_CACHE> image_pool;
    queue<VINS_DATA_CACHE> vins_pool;
    IMG_DATA_CACHE image_data_cache;
    cv::Mat lateast_equa;
    cv::Mat lateast_image;
    Vector3f lateast_P;
    Matrix3f lateast_R;
    cv::Mat pnp_image;
    Vector3d pnp_P;
    Matrix3d pnp_R;


    inline static double timeStampToSec (long timeStamp) { return timeStamp / 1000000000.0; };
    static NSTimeInterval systemUptime();
    NSTimeInterval getLateast_imu_time() const;

    void testMethod(){ LOGI("VINSController Testmethod is working."); }

    // Setup controller's VINS, create threads.
    void viewDidLoad();

    // Check device model
    DeviceType deviceName();

    void processImage(cv::Mat& image, double timeStamp, bool isScreenRotated);

    // From buf frame by frame obtain img_msg with all imu_msg-s up to the img frame.
    std::vector<std::pair<std::vector<ImuConstPtr>, ImgConstPtr>> getMeasurements();

    // Used for feature tracking, returns and removes all imu_msg from the local imu buffer
    // that are in between the last and the current header (timeStamp)
    vector<IMU_MSG_LOCAL> getImuMeasurements(double header);

    // Send imu_msg to vins, then call vins.processIMU()
    void send_imu(const ImuConstPtr &imu_msg);

    bool mainLoop_isCancelled = false;
    // VINS thread: this thread tightly fuses the visual measurements and imu data
    // and solves pose, velocity, IMU bias, 3D feature for all frame in WINNDOW
    // If the newest frame is keyframe, then push it into keyframe database
    void run();

    int kf_global_index;
    bool start_global_optimization = false;
    void process();

    bool loop_thread_isCancelled = false;
    // Loop detection thread: this thread detect loop for newest keyframe and retrieve features
    void loopDetectionLoop();

    bool globalLoopThread_isCancelled = false;
    // GLobal Pose graph thread: optimize global pose graph based on realative pose from vins and update the keyframe database
    void globalPoseGraphOptimizationLoop();

    /*
     * Z^
     * |   /Y
     * |  /
     * | /
     * |/
     * --------->X
     * IMU data process and interpolation 
     */
    bool imuDataFinished = false;
    bool vinsDataFinished = false;
    shared_ptr<IMU_MSG> cur_acc = shared_ptr<IMU_MSG>(new IMU_MSG()); // shared_ptr<IMU_MSG> cur_acc(new IMU_MSG());
    vector<IMU_MSG> gyro_buf;  // buf for Interpolation
    void imuStartUpdate();
    void imuStopUpdate();

    /** UI View Controler */
    void showInputView();
    void showOutputImage(cv::Mat* image) {}

    /** UI Button Controler */
    // TODO: UI Interaction
    // switch UI between graph(VINS) and camera(AR)
    void switchUI(bool isChecked);
    
    // loop button
    void loopButtonPressed(bool isChecked);




    /** About record and playback data for debug */
    bool saveData_isCancelled = false;
    void saveDataLoop();

    void tapSaveImageToIphone(cv::Mat* image) {
        // TODO: save image to Album
        // UIImageWriteToSavedPhotosAlbum(image, self, @selector(image:didFinishSavingWithError:contextInfo:), nil);
    }

//TODO: is this even used anywhere? -> seems to be a callback to process error while saving to filesystem
    /*
    void image:(UIImage *)image didFinishSavingWithError:(NSError *)error contextInfo:(void *)contextInfo{
//- (void)image:(UIImage *)image didFinishSavingWithError:(NSError *)error contextInfo:(void *)contextInfo{

            if (error == nil) {
                __android_log_print(ANDROID_LOG_INFO, APPNAME, "save access");
            }else{
                __android_log_print(ANDROID_LOG_INFO, APPNAME, "save failed");
            }
    }
    */

//TODO: do directory management work
    /*
    void checkDirectoryPath:(unsigned long)index withObject:(NSString*)directoryPath {
//- (void)checkDirectoryPath:(unsigned long)index withObject:(NSString*)directoryPath
//{
        //delete already exist directory first time
        NSError *error;
        if (index == 0 && [[NSFileManager defaultManager] fileExistsAtPath:directoryPath])	//Does directory exist?
        {
            if (![[NSFileManager defaultManager] removeItemAtPath:directoryPath error:&error])	//Delete it
            {
                __android_log_print(ANDROID_LOG_INFO, APPNAME, "Delete directory error: %@", error);
            }
        }

        //creat file directory if it does not exist
        if (![[NSFileManager defaultManager] fileExistsAtPath:directoryPath])
        {
            __android_log_print(ANDROID_LOG_INFO, APPNAME, "directory does not exist");
            if (![[NSFileManager defaultManager] createDirectoryAtPath:directoryPath
            withIntermediateDirectories:NO
            attributes:nil
            error:&error])
            {
                __android_log_print(ANDROID_LOG_INFO, APPNAME, "Create directory error: %@", error);
            }
        }
    }
    */

    void recordImu() {
        // TODO: write imuDataBuf to file
        /*
        NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        NSString *documentsPath = [paths objectAtIndex:0];
        NSString *filePath = [documentsPath stringByAppendingPathComponent:@"IMU"]; //Add the file name

        [imuDataBuf writeToFile:filePath atomically:YES];
        //[msgData writeToFile:filePath atomically:YES];
        */
    }

    void recordVins() {
        // TODO: write vinsDataBuf to file
        /*
        NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        NSString *documentsPath = [paths objectAtIndex:0];
        NSString *filePath = [documentsPath stringByAppendingPathComponent:@"VINS"]; //Add the file name

        [vinsDataBuf writeToFile:filePath atomically:YES];
        //[msgData writeToFile:filePath atomically:YES];
        */
    }

    void recordImageTime(IMG_DATA& image_data) {
        // TODO: write image_data.header (time) to file
        /*
        double time = image_data.header;
        NSData *msgData = [NSData dataWithBytes:&time length:sizeof(time)];
        NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        NSString *documentsPath = [[paths objectAtIndex:0] stringByAppendingPathComponent:@"IMAGE_TIME"];; //Get the docs directory

        [self checkDirectoryPath:imageDataIndex withObject:documentsPath];

        NSString *filename = [NSString stringWithFormat:@"%lu", imageDataIndex];
        NSString *filePath = [documentsPath stringByAppendingPathComponent:filename]; //Add the file name

        [msgData writeToFile:filePath atomically:YES];
        */
    }

    void recordImage(IMG_DATA& image_data) {
        // TODO: write image_data.image to file
        /*
        NSData *msgData = UIImagePNGRepresentation(image_data.image);
        //NSData *msgData = [NSData dataWithBytes:&image_data length:sizeof(image_data)];
        NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        NSString *documentsPath = [[paths objectAtIndex:0] stringByAppendingPathComponent:@"IMAGE"];; //Get the docs directory

        [self checkDirectoryPath:imageDataIndex withObject:documentsPath];

        NSString *filename = [NSString stringWithFormat:@"%lu", imageDataIndex];
        NSString *filePath = [documentsPath stringByAppendingPathComponent:filename]; //Add the file name

        [msgData writeToFile:filePath atomically:YES];
        */
    }

    // tries to read image time from file and writes it into property imgData.header
    bool readImageTime(unsigned long index) {
        return false;
        // TODO: implement
        /*
        bool file_exist;
        NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        NSString *documentsPath = [[paths objectAtIndex:0] stringByAppendingPathComponent:@"IMAGE_TIME"]; //Get the docs directory
        NSString *filename = [NSString stringWithFormat:@"%lu", index];
        NSString *filePath = [documentsPath stringByAppendingPathComponent:filename]; //Add the file name

        //check file exists
        if ([[NSFileManager defaultManager] fileExistsAtPath:filePath])
        {
            NSData *file1 = [[NSData alloc] initWithContentsOfFile:filePath];
            if (file1)
            {
                double time;
                [file1 getBytes:&time length:sizeof(time)];
                imgData.header = time;
            }
            file_exist = true;
        }
        else
        {
            file_exist = false;
            //__android_log_print(ANDROID_LOG_INFO, APPNAME, "File does not exist");
        }
        return file_exist;
        */
    }

    // tries to read image from file and writes it into property imgData.image
    bool readImage(unsigned long index) {
        return false;
        // TODO: implement
        /*
        bool file_exist;
        NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        NSString *documentsPath = [[paths objectAtIndex:0] stringByAppendingPathComponent:@"IMAGE"]; //Get the docs directory
        NSString *filename = [NSString stringWithFormat:@"%lu", index];
        NSString *filePath = [documentsPath stringByAppendingPathComponent:filename]; //Add the file name

        //check file exists
        if ([[NSFileManager defaultManager] fileExistsAtPath:filePath])
        {
            NSData *pngData = [NSData dataWithContentsOfFile:filePath];
            imgData.image = [UIImage imageWithData:pngData];
            file_exist = true;
        }
        else
        {
            file_exist = false;
            //__android_log_print(ANDROID_LOG_INFO, APPNAME, "File does not exist");
        }
        return file_exist;
        */
    }




    // TODO: call when app closes
    void viewDidDisappear() {
        if (isCapturing) {
            // TODO: stop the camera [videoCamera stop];
        }
        mainLoop_isCancelled = true;
        // TODO: where does it get started? [draw cancel];
#ifdef LOOP_CLOSURE
        loop_thread_isCancelled = true; // [loopDetectThread cancel];
#endif
    }

    void viewDidUnload() {}

    void dealloc() {}

    double autoClipLimit(Mat mat);
};


#endif //VINS_MOBILE_ANDROIDPORT_VIEWCONTROLLER_H
